#!/usr/bin/env python
# -*- coding: utf-8 -*- 
from __future__ import print_function

"""
Empirical XMM background model.

Written by Zhu Liu, adapted by Torben Simm and Johannes Buchner
(C) 2013-2016

For example usage, see examples/sherpa/background/xmm/fit.py
"""	

import os
import numpy

if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat
	from sherpa.models.parameter import Parameter
	from sherpa.models import ArithmeticModel, CompositeModel
	from sherpa.astro.ui import *
	from sherpa.astro.instrument import RSPModelNoPHA, RMFModelNoPHA
else:
	CompositeModel, ArithmeticModel = object, object

print("""

Using XMM empirical background model originally by Richard Sturm.
Please reference Maggi P., et al., 2014, A&A, 561, AA76.

""")


def get_embedded_file(filename):
    """
    Gets the path of a file in the same folder as this script
    """
    return os.path.join(os.path.dirname(__file__), filename)

def get_pn_bkg_model(i, galabs, fit=False):
#=================================================================
# Parameters:
# i = sherpa ID of data set  
# galabs = name of model for galactic absorption
# pnbrsp = response model of bkg data set 
# fit: True = fit bkg spectrum; False = just set_bkg_full_model
# returns unit response for bkg model
#=================================================================
    #get instrument response model for bkg spectrum
    if get_rmf(i).energ_lo[0] == 0: get_rmf(i).energ_lo[0] = 0.001
    if get_arf(i).energ_lo[0] == 0: get_arf(i).energ_lo[0] = 0.001 
    pnbrsp = get_response(i,bkg_id=1)
    pnscale = get_bkg_scale(i) #get background scaling factor 

    #create unit response
    dia_pn_rmf=get_embedded_file('pn_dia.rmf')
    dia_pn_arf=get_embedded_file('pn_dia.arf')
    copy_data(i,1000+2) 	
    load_bkg_rmf(1000+2, dia_pn_rmf) #load diagonal bkg matrices
    load_bkg_arf(1000+2, dia_pn_arf)
    pnbunitrsp = get_response(1000+2, bkg_id=1)	
    delete_data(1000+2)

    #gaussian line center energy, line width, and initial normalization; for *PN background*
    pncenters = [1.49165, 1.49165, 4.53177, 5.42516, 6.38155, 7.48675, 8.04087, 8.04087, 8.60924, 8.89395, 9.56160]
    pnlinewidth = [5.73813E-02, 3.63469E-02, 6.10487E-02, 7.08380E-02, 9.59053E-02, 6.52422E-02, 9.48594E-02, 6.26174E-05, 0.120893, 0.114254, 0.108717]
    pnlinenorm = [7.81356E-03, 3.96601E-03, 7.30727E-04, 4.96413E-04, 5.31295E-9, 6.84796E-04, 3.01564E-02, 1.41847E-04, 8.87887E-03, 5.75592E-03, 1.71367E-03]

    pnbkgcons, pnbkgspline1, pnbkgexpdec, pnbkgsmedge1, pnbkgsmedge2, pnbkgspline2, pnbkginspl, pnbkgline1, pnbkgline2, pnbkgline3, pnbkgline4, pnbkgline5, pnbkgline6, pnbkgline7, pnbkgline8, pnbkgline9, pnbkgline10, pnbkgline11, pnbkgpl, pnbkgapec, pnbkglcapec, = (xsconstant.pncons, xsspline.pnspline1, xsexpdec.pnexpdec, xssmedge.pnsmedge1, xssmedge.pnsmedge2, xsspline.pnspline2, xspowerlaw.pnbkpl, xsgaussian.pngau1, xsgaussian.pngau2, xsgaussian.pngau3, xsgaussian.pngau4, xsgaussian.pngau5, xsgaussian.pngau6, xsgaussian.pngau7, xsgaussian.pngau8, xsgaussian.pngau9, xsgaussian.pngau10, xsgaussian.pngau11, xspowerlaw.pnpnexpl, xsapec.pnapec, xsapec.pnlcapec)

    pnlines = [pnbkgline1, pnbkgline2, pnbkgline3, pnbkgline4, pnbkgline5, pnbkgline6, pnbkgline7, pnbkgline8, pnbkgline9, pnbkgline10, pnbkgline11]

    pnfixwid = [pnbkgline2, pnbkgline3, pnbkgline4, pnbkgline5, pnbkgline6, pnbkgline8, pnbkgline11]
    pnfree = [pnbkgline1, pnbkgline7, pnbkgline9, pnbkgline10]

    #define PN background model
    pn_bkg = pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
	     pnbkgline1 + pnbkgline2 + pnbkgline3 + pnbkgline4 + pnbkgline5 + pnbkgline6 + pnbkgline7 + pnbkgline8 + 
	     pnbkgline9 + pnbkgline10 + pnbkgline11))) + pnbrsp(galabs*(pnbkgpl+pnbkgapec)+pnbkglcapec)

    for l, c in zip(pnlines, pncenters):
        l.LineE = c
        l.LineE.min = c - 0.05
        l.LineE.max = c + 0.05

    pnbkgline2.LineE = pnbkgline1.LineE
    pnbkgline8.LineE = pnbkgline7.LineE 

    for l, s in zip(pnlines, pnlinewidth):
        l.Sigma = s
        
    for l in pnfree:
        l.Sigma.min = 1E-5
        l.Sigma.max = 0.2
			
    for l in pnfixwid:
        l.Sigma.freeze()
			
    for l, n in zip(pnlines, pnlinenorm):
        l.norm = n
        l.norm.min = 1E-10
        l.norm.max = 1E10

    # Scaling constant
    pnbkgcons.factor = 1.0
    pnbkgcons.factor.freeze()
    
    # thermal radiation
    pnbkgapec.kT = 0.286928
    pnbkgapec.kT.min = 0.008
    pnbkgapec.kT.max = 64
    pnbkgapec.Abundanc = 1.0
    pnbkgapec.Abundanc.freeze()
    pnbkgapec.Redshift = 0.0
    pnbkgapec.Redshift.freeze()
    pnbkgapec.norm = 5.58410E-05
    pnbkgapec.norm.min = 1e-10
    pnbkgapec.norm.max = 1e10
    
    # local thermal radiation
    pnbkglcapec.kT = 0.1
    pnbkglcapec.kT.freeze()
    pnbkglcapec.Abundanc = 1.0
    pnbkglcapec.Abundanc.freeze()
    pnbkglcapec.Redshift = 0.0
    pnbkglcapec.Redshift.freeze()
    pnbkglcapec.norm = 3.89164E-05
    pnbkglcapec.norm.min = 1e-10
    pnbkglcapec.norm.max = 1e10
    
    # exponential decay
    pnbkgexpdec.factor = 44.3418
    pnbkgexpdec.factor.min = 0
    pnbkgexpdec.factor.max = 100
    pnbkgexpdec.norm = 6830.89
    pnbkgexpdec.norm.freeze()
    
    # Smear function
    pnbkgsmedge1.edgeE = 0.538408
    pnbkgsmedge1.edgeE.freeze()
    pnbkgsmedge1.MaxTau = 1.40238
    pnbkgsmedge1.MaxTau.min = 0 
    pnbkgsmedge1.MaxTau.max = 10
    pnbkgsmedge1.index = -2.67000
    pnbkgsmedge1.index.freeze()
    pnbkgsmedge1.width = 0.313365
    pnbkgsmedge1.width.min = 0.01
    pnbkgsmedge1.width.max = 100
    
    pnbkgsmedge2.edgeE = 1.38826
    pnbkgsmedge2.edgeE.freeze()
    pnbkgsmedge2.MaxTau.min = 0 
    pnbkgsmedge2.MaxTau.max = 10
    pnbkgsmedge2.MaxTau = 9.37167
    pnbkgsmedge2.index = -2.67000
    pnbkgsmedge2.index.freeze()
    pnbkgsmedge2.width = 5.7642
    pnbkgsmedge2.width.min = 0.01
    pnbkgsmedge2.width.max = 100
    	
    # Spline funtion
    pnbkgspline1.Estart = 0.200000
    pnbkgspline1.Estart.freeze()
    pnbkgspline1.Ystart = -1.31506
    pnbkgspline1.Ystart.min = -1e+6
    pnbkgspline1.Ystart.max = 1e+6
    pnbkgspline1.Yend = 1064.16
    pnbkgspline1.Yend.min = -1e+6
    pnbkgspline1.Yend.max = 1e+6
    pnbkgspline1.YPstart = -106.183
    pnbkgspline1.YPstart.min = -1e+6
    pnbkgspline1.YPstart.max = 1e+6
    pnbkgspline1.YPend = -366.092
    pnbkgspline1.YPend.min = -1e+6
    pnbkgspline1.YPend.max = 1e6
    pnbkgspline1.Eend = 1.74715
    pnbkgspline1.Eend.min = 0
    pnbkgspline1.Eend.max = 100
    
    pnbkgspline2.Estart = 3.29056
    pnbkgspline2.Ystart = 1.00643
    pnbkgspline2.Ystart.min = -1e+6
    pnbkgspline2.Ystart.max = 1e+6
    pnbkgspline2.Yend = 0.887026
    pnbkgspline2.Yend.min = -1e+6
    pnbkgspline2.Yend.max = 1e+6
    pnbkgspline2.YPstart = -0.278401
    pnbkgspline2.YPstart.min = -1e+6
    pnbkgspline2.YPstart.max = 1e+6
    pnbkgspline2.YPend = 4.84809E-03
    pnbkgspline2.YPend.min = -1e+6
    pnbkgspline2.YPend.max = 1e6
    pnbkgspline2.Eend = 7.32701
    pnbkgspline2.Eend.min = 0
    pnbkgspline2.Eend.max = 100
    
    # Background power law
    pnbkginspl.PhoIndex = 0.279
    pnbkginspl.PhoIndex.min = -2
    pnbkginspl.PhoIndex.max = 9
    pnbkginspl.norm = 8.23614E-03
    pnbkginspl.norm.min = 1E-10
    pnbkginspl.norm.max = 1E6
    
    # Extragalactic background
    pnbkgpl.PhoIndex = 1.46
    pnbkgpl.PhoIndex.freeze()
    pnbkgpl.norm = 1.25288E-05
    pnbkgpl.norm.min = 1e-10
    pnbkgpl.norm.max = 1e3

    if(fit): #fit bkg 
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl))))
        print('Fitting (1/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
					pnbkgline2))))
        print('Fitting (2/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
					pnbkgline1 + pnbkgline2))))
        print('Fitting (3/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
					pnbkgline1 + pnbkgline2 + pnbkgline7 + pnbkgline8))))
        print('Fitting (4/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
					pnbkgline1 + pnbkgline2 + pnbkgline7 + pnbkgline8 + pnbkgline9 + pnbkgline10))))
        print('Fitting (5/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
        				pnbkgline1 + pnbkgline2 + pnbkgline3 + pnbkgline4 + pnbkgline5 + pnbkgline6 + pnbkgline7 + pnbkgline8 + pnbkgline9 + 
					pnbkgline10 + pnbkgline11))))
        print('Fitting (6/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
					pnbkgline1 + pnbkgline2 + pnbkgline3 + pnbkgline4 + pnbkgline5 + pnbkgline6 + pnbkgline7 + pnbkgline8 + pnbkgline9 + 
					pnbkgline10 + pnbkgline11))) + pnbrsp(galabs*(pnbkgapec)))
        print('Fitting (7/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
					pnbkgline1 + pnbkgline2 + pnbkgline3 + pnbkgline4 + pnbkgline5 + pnbkgline6 + pnbkgline7 + pnbkgline8 + pnbkgline9 + 
					pnbkgline10 + pnbkgline11))) + pnbrsp(galabs*(pnbkgapec+pnbkgpl)))
        print('Fitting (8/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
					pnbkgline1 + pnbkgline2 + pnbkgline3 + pnbkgline4 + pnbkgline5 + pnbkgline6 + pnbkgline7 + pnbkgline8 + pnbkgline9 + 
					pnbkgline10 + pnbkgline11))) + pnbrsp(galabs*(pnbkgapec+pnbkgpl)+pnbkglcapec))
        fit_bkg(i)
        
        freeze(pnbkgcons, pnbkgspline1, pnbkgexpdec, pnbkgsmedge1, pnbkgsmedge2, pnbkgspline2, pnbkginspl, pnbkgline1, pnbkgline2, 
					pnbkgline3, pnbkgline4, pnbkgline5, pnbkgline6, pnbkgline7, pnbkgline8, pnbkgline9, pnbkgline10, pnbkgline11, 
					galabs, pnbkgapec, pnbkgpl, pnbkglcapec)
        print(' ')
        print('PN background model set up and fitted')
        print('Please double-check that it is a good fit')
        print(' ')
    else:
        set_bkg_full_model(i,pnbunitrsp(pnbkgcons*(pnbkgspline1*pnbkgexpdec + pnbkgsmedge1*pnbkgsmedge2*(pnbkgspline2*pnbkginspl + 
	       pnbkgline1 + pnbkgline2 + pnbkgline3 + pnbkgline4 + pnbkgline5 + pnbkgline6 + pnbkgline7 + pnbkgline8 + pnbkgline9 + 
	       pnbkgline10 + pnbkgline11))) + pnbrsp(galabs*(pnbkgapec+pnbkgpl)+pnbkglcapec))
        print(' ')
        print('PN background model set up')
        print(' ')
    return pnscale*(pnbunitrsp(pncons*(pnspline1*pnexpdec + pnsmedge1*pnsmedge2*(pnspline2*pnbkpl + 
			pngau1 + pngau2 + pngau3 + pngau4 + pngau5 + pngau6 + pngau7 + pngau8 + pngau9 + 
			pngau10 + pngau11))) + pnbrsp(galabs*(pnapec+pnpnexpl)+pnlcapec))


def get_mos_bkg_model(i, galabs, fit=False):
#=================================================================
# Parameters:
# i = sherpa ID of data set  
# galabs = name of model for galactic absorption
# fit: True = fit bkg spectrum; False = just set_bkg_full_model
# returns unit response for bkg model
#=================================================================
    #get instrument response model for bkg spectrum
    if get_rmf(i).energ_lo[0] == 0: get_rmf(i).energ_lo[0] = 0.001
    if get_arf(i).energ_lo[0] == 0: get_arf(i).energ_lo[0] = 0.001 
    mosbrsp = get_response(i,bkg_id=1)
    mosscale = get_bkg_scale(i) #get background scaling factor 

    #create unit response
    dia_mos_rmf=get_embedded_file('mos_dia.rmf')
    dia_mos_arf=get_embedded_file('mos_dia.arf')
    copy_data(i,1000+2) 	
    load_bkg_rmf(1000+2, dia_mos_rmf) #load diagonal bkg matrices
    load_bkg_arf(1000+2, dia_mos_arf)
    mosbunitrsp = get_response(1000+2, bkg_id=1)	
    delete_data(1000+2)

    #gaussian line center energy, line width, and initial normalization; for *MOS background*
    moscenters = [1.48600, 1.48700, 1.74000, 5.41000, 5.89500, 6.42000, 9.71000]
    moslinewidth = [3.84602E-02, 0.165816, 3.54985E-02, 9.77018E-02, 7.45076E-02, 7.42365E-02, 9.04855E-02]
    moslinenorm = [9.93119E-03, 1.67028E-03, 1.75461E-03, 2.86358E-04, 2.07525E-04, 3.07555E-04, 4.58115E-04]
    
    # model component prefix
    mos = 'mos%s' % i

    mosbkgcons, mosbkgsmedge, mosbkgspline, mosbkgbknpl, mosbkgline1, mosbkgline2, mosbkgline3, mosbkgline4, mosbkgline5, mosbkgline6, mosbkgline7, mosbkgpl, mosbkgapec, mosbkglcapec = (xsconstant(mos+'cons'), xssmedge(mos+'smedge'), xsspline(mos+'spline'), xsbknpower(mos+'bknpl'), xsgaussian(mos+'gau1'), xsgaussian(mos+'gau2'), xsgaussian(mos+'gau3'), xsgaussian(mos+'gau4'), xsgaussian(mos+'gau5'), xsgaussian(mos+'gau6'), xsgaussian(mos+'gau7'), xspowerlaw(mos+'expl'), xsapec(mos+'apec'), xsapec(mos+'lcapec'))

    moslines = [mosbkgline1, mosbkgline2, mosbkgline3, mosbkgline4, mosbkgline5, mosbkgline6, mosbkgline7] 
    mos_bkg = 'mos_bkg_model'  

    #define MOS background model
    mos_bkg = mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + mosbkgline2 + mosbkgline3 + 
	      mosbkgline4 + mosbkgline5 + mosbkgline6 + mosbkgline7) + mosbrsp(galabs*(mosbkgpl + mosbkgapec) + mosbkglcapec)

    for l, c in zip(moslines, moscenters):
        l.LineE = c
        l.LineE.min = c - 0.05
        l.LineE.max = c + 0.05
			
    for l, s in zip(moslines, moslinewidth):
        l.Sigma = s
        l.Sigma.min = 1E-4
        l.Sigma.max = 0.2
    mosbkgline1.Sigma.min = 1E-4
    mosbkgline1.Sigma.max = 1E-1

    for l, n in zip(moslines, moslinenorm):
        l.norm = n
        l.norm.min = 1E-10
        l.norm.max = 1E10

    # Constant factor
    mosbkgcons.factor = 1.0
    mosbkgcons.factor.freeze()
    
    # Smear function
    mosbkgsmedge.edgeE = 0.538408
    mosbkgsmedge.edgeE.freeze()
    mosbkgsmedge.MaxTau = 0.246633
    mosbkgsmedge.MaxTau.min = 0.0
    mosbkgsmedge.MaxTau.max = 10.0
    mosbkgsmedge.index = -2.67
    mosbkgsmedge.index.freeze()
    mosbkgsmedge.width = 1E-02 
    mosbkgsmedge.width.min = 1E-02
    mosbkgsmedge.width.max = 1E2
    
    # Spline funtion
    mosbkgspline.Estart = 3.08175
    mosbkgspline.Ystart = 1.00984
    mosbkgspline.Yend = 1.99144
    mosbkgspline.YPstart = -2.90195e-02
    mosbkgspline.YPend = 5.49102e-02
    mosbkgspline.Estart.freeze()
    mosbkgspline.Ystart.freeze()
    mosbkgspline.Yend.freeze()
    mosbkgspline.YPstart.freeze()
    mosbkgspline.YPend.freeze()
    mosbkgspline.Eend = 13.6492
    mosbkgspline.Eend.min = 0
    mosbkgspline.Eend.max = 100
    
    # Broken power law
    mosbkgbknpl.PhoIndx1 = 1.48636
    mosbkgbknpl.PhoIndx1.min = -2
    mosbkgbknpl.PhoIndx1.max = 9
    mosbkgbknpl.BreakE = 0.415173
    mosbkgbknpl.PhoIndx2 = 0.315615
    mosbkgbknpl.BreakE.freeze()
    mosbkgbknpl.PhoIndx2.freeze()
    mosbkgbknpl.norm = 2.90071E-03
    mosbkgbknpl.norm.min = 1E-10
    mosbkgbknpl.norm.max = 1E10
     
    # Extragalactic background
    mosbkgpl.PhoIndex = 1.46
    mosbkgpl.PhoIndex.freeze()
    mosbkgpl.norm = 4.58115E-04
    mosbkgpl.norm.min = 1E-10
    mosbkgpl.norm.max = 1E10
    
    # thermal radiation
    mosbkgapec.kT = 0.286928
    mosbkgapec.kT.min = 0.008
    mosbkgapec.kT.max = 64
    mosbkgapec.Abundanc = 1.0
    mosbkgapec.Abundanc.freeze()
    mosbkgapec.Redshift = 0.0
    mosbkgapec.Redshift.freeze()
    mosbkgapec.norm = 5.58410E-05
    mosbkgapec.norm.min = 1e-10
    mosbkgapec.norm.max = 1e10
    
    # local thermal radiation
    mosbkglcapec.kT = 0.1
    mosbkglcapec.kT.freeze()
    mosbkglcapec.Abundanc = 1.0
    mosbkglcapec.Abundanc.freeze()
    mosbkglcapec.Redshift = 0.0
    mosbkglcapec.Redshift.freeze()
    mosbkglcapec.norm = 3.89164E-05
    mosbkglcapec.norm.min = 1e-10
    mosbkglcapec.norm.max = 1e10

    if(fit): #fit bkg 
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl))
        print('Fitting (1/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline2))
        print('Fitting (2/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline2 + 
					mosbkgline3))
        print('Fitting (3/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + 
					mosbkgline2 + mosbkgline3))
        print('Fitting (4/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + 
					mosbkgline2 + mosbkgline3 + mosbkgline4 + mosbkgline5 + mosbkgline6 + mosbkgline7))
        print('Fitting (5/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + 
					mosbkgline2 + mosbkgline3 + mosbkgline4 + mosbkgline5 + mosbkgline6 + mosbkgline7) + 
					mosbrsp(galabs*(mosbkgapec)))
        print('Fitting (6/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + 
					mosbkgline2 + mosbkgline3 + mosbkgline4 + mosbkgline5 + mosbkgline6 + mosbkgline7) + 
					mosbrsp(galabs*(mosbkgapec+mosbkgpl)))
        print('Fitting (7/8)...')
        fit_bkg(i)
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + 
					mosbkgline2 + mosbkgline3 + mosbkgline4 + mosbkgline5 + mosbkgline6 + mosbkgline7) + 
					mosbrsp(galabs*(mosbkgapec+mosbkgpl)+mosbkglcapec))
        print('Fitting (8/8)...')
        fit_bkg(i)
        freeze(mosbkgcons, mosbkgsmedge, mosbkgspline, mosbkgbknpl, mosbkgline1, mosbkgline2, mosbkgline3, 
					mosbkgline4, mosbkgline5, mosbkgline6, mosbkgline7, mosbkgpl, mosbkgapec, mosbkglcapec)
        print(' ')
        print('MOS background model set up and fitted')
        print('Please double-check that it is a good fit')
        print(' ')
    else:
        set_bkg_full_model(i,mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + 
					mosbkgline2 + mosbkgline3 + mosbkgline4 + mosbkgline5 + mosbkgline6 + mosbkgline7) + 
					mosbrsp(galabs*(mosbkgapec+mosbkgpl)+mosbkglcapec))
        print(' ')
        print('MOS background model set up')
        print(' ')
    
    return mosscale*(mosbunitrsp(mosbkgcons*mosbkgsmedge*mosbkgspline*mosbkgbknpl + mosbkgline1 + 
        mosbkgline2 + mosbkgline3 + mosbkgline4 + mosbkgline5 + mosbkgline6 + mosbkgline7) + 
        mosbrsp(galabs*(mosbkgapec+mosbkgpl)+mosbkglcapec))
    
    #return mosscale*(mosbunitrsp(moscons*mossmedge*mosspline*mosbknpl + mosgau1 + mosgau2 + mosgau3 + 
    #    		mosgau4 + mosgau5 + mosgau6 + mosgau7) + mosbrsp(galabs*(mosapec+mosexpl)+moslcapec))

def get_mos_bkg_model_cached(i, galabs):
	filename = get_bkg(i).name + '.bkgpars'
	if os.path.exists(filename):
		bkgmodel = get_mos_bkg_model(i, galabs, fit=False)
		for p, v in zip(bkgmodel.pars, numpy.loadtxt(filename)):
			p.val = v
	else:
		bkgmodel = get_mos_bkg_model(i, galabs, fit=True)
		numpy.savetxt(filename, [p.val for p in bkgmodel.pars])
	for p in bkgmodel.pars:
		p.freeze()
	return bkgmodel
	
def get_pn_bkg_model_cached(i, galabs):
	filename = get_bkg(i).name + '.bkgpars'
	if os.path.exists(filename):
		bkgmodel = get_pn_bkg_model(i, galabs, fit=False)
		for p, v in zip(bkgmodel.pars, numpy.loadtxt(filename)):
			p.val = v
	else:
		bkgmodel = get_pn_bkg_model(i, galabs, fit=True)
		numpy.savetxt(filename, [p.val for p in bkgmodel.pars])
	for p in bkgmodel.pars:
		p.freeze()
	return bkgmodel



#!/usr/bin/env python
# -*- coding: utf-8 -*- 
"""
BXA (Bayesian X-ray Analysis) for Sherpa

Copyright: Johannes Buchner (C) 2013-2025
"""

import os

if 'MAKESPHINXDOC' not in os.environ:
	from sherpa.models import ArithmeticModel, CompositeModel
	from sherpa.models.parameter import Parameter
else:
	ArithmeticModel = object
	CompositeModel = list
	Parameter = None


class VariableCachedModel(CompositeModel, ArithmeticModel):
	"""Wrapper that caches the most recent model call."""
	def __init__(self, othermodel):
		"""*othermodel* can be any sherpa model"""
		self.othermodel = othermodel
		self.cache = None
		self.lastp = None
		print('calling CompositeModel...')
		CompositeModel.__init__(self, name='cached(%s)' % othermodel.name, parts=(othermodel,))
	
	def calc(self, p, left, right, *args, **kwargs):
		if self.cache is None or self.lastp != p:
			self.cache = self.othermodel.calc(p, left, right, *args, **kwargs)
			self.lastp = p
		return self.cache

	def startup(self):
		self.othermodel.startup()
		CompositeModel.startup(self)
	
	def teardown(self):
		self.othermodel.teardown()
		CompositeModel.teardown(self)

	def guess(self, dep, *args, **kwargs):
		self.othermodel.guess(dep, *args, **kwargs)
		CompositeModel.guess(self, dep, *args, **kwargs)

class CachedModel(CompositeModel, ArithmeticModel):
	"""Wrapper that caches the first model call forever."""
	def __init__(self, othermodel):
		"""*othermodel* can be any sherpa model"""
		self.othermodel = othermodel
		self.cache = None
		self.relnorm = Parameter('cache', 'relnorm', 1, 0, 1e10, 0, 1e10)
		CompositeModel.__init__(self, name='cached(%s)' % othermodel.name, parts=(othermodel,))
	
	def calc(self, *args, **kwargs):
		if self.cache is None:
			print('   computing cached model ... ')
			self.cache = self.othermodel.calc(*args, **kwargs)
		return self.cache * self.relnorm.val

	def startup(self):
		self.othermodel.startup()
		CompositeModel.startup(self)
	
	def teardown(self):
		self.othermodel.teardown()
		CompositeModel.teardown(self)

	def guess(self, dep, *args, **kwargs):
		self.othermodel.guess(dep, *args, **kwargs)
		CompositeModel.guess(self, dep, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*- 

"""
Binning routines for plotting
"""
import numpy
import scipy.special, scipy.stats
from . import gof
import tqdm


def group_adapt(xdata, ydata, xlo, xhi, nmin = 20):
	"""
	Adaptive grouping into nmin count bins.
	
	Args:
		:param xdata: bin center
		:param ydata: count values
		:param xlo: bin lower edge
		:param xhi: bin upper edge
		:param nmin: desired minimum number of counts in each bin
	
	Returns:
		sequence of (lower edge, upper edge, number of counts) tuples
	"""
	i = 0
	while i < len(xlo):
		for j in range(i, len(xlo)):
			# [i:j] (with j)
			xmask = numpy.logical_and(xdata >= xlo[i], xdata < xhi[j])
			
			if ydata[xmask].sum() >= nmin or j + 1 >= len(xlo):
				yield (xlo[i], xhi[j], ydata[xmask].sum())
				#print '  groups', i,j
				break
		i = j + 1


def binning(outputfiles_basename, bins, widths, data, models, nmin=20):
	"""Bins the data for plotting and checks the model.

	Using the gof module, computes a Poisson goodness-of-fit range,
	i.e. ranges where the model must lie. This is done for multiple
	binning sizes simultaneously.

	:param outputfiles_basename: not used.
	:param bins: bin location from Plot.x()
	:param widths: bin width from Plot.xErr()
	:param data: counts per bin width from Plot.y() with Plot.background = True and Plot('counts')
	:param model: counts per bin width predicted by the model, from Plot.model()
	:param nmin: number of counts per bin to use for rebinning.
	
	Returns
	--------
	
	* marked_binned: data points binned to contain `nmin` counts
	  a sequence ready to be passed to matplotlib.pyplot.errorbar
	* modelrange: range allowed by the data
	  ready to be passed to matplotlib.pyplot.fill_between
	* and statistics (GoF measure)
	
	"""
	
	xdata = bins
	xlo = bins - widths
	xhi = bins + widths
	# convert from densities to counts
	ydata = numpy.rint(data * widths * 2)
	models = models * widths * 2
	
	best_gof = None
	best_gof_stats = None
	data = None
	
	grouped_data = list(group_adapt(xdata, ydata, xlo, xhi, nmin=nmin))
	data = numpy.array([ydata[numpy.logical_and(xdata >= i, xdata < j)].sum() for i, j in zip(xlo, xhi)])
	
	component = models[:,0,:]
	for i, counts_predicted in enumerate(tqdm.tqdm(component)):
		modelrange_low, modelrange_high = gof.calc_models_range(data)
	
		stats = gof.calc_multigof(data, counts_predicted)
		curgof = -numpy.log10(
			numpy.min([stats[stats[:,0] == n][:,2].min() * (stats[:,0] == n).sum() 
				for n in sorted(set(stats[:,0]))]) + 1e-300)
		
		if best_gof is None or curgof < best_gof:
			best_gof = curgof
			best_gof_stats = stats
		if i > 100:
			break

	# check if we can reproduce the data
	curgof = best_gof
	stats = best_gof_stats
	
	data_gofp = [numpy.nan] * len(grouped_data)
	for n in numpy.unique(stats[:,0].astype(int)):
		# find the worst case for this level and each datapoint
		#exp(numpy.log(stats[stats[:,0] == n][:,2]).sum()) * (stats[:,0] == n).sum() 
		#		for n in sorted(set(stats[:,0]))]))
		nstats = stats[stats[:,0] == n]
		pxlo = xlo[(nstats[:,1] * n).astype(int)]
		pxhi = numpy.asarray(pxlo[1:].tolist() + [xdata.max()])
	
		# so far so good.
		# mark data points that have not been achieved
		#print zip(pxlo, chi2min)
		for i, (xloi, xhii, ydatai) in enumerate(grouped_data):
			# select p values that intersect
			mask = numpy.logical_and(pxlo < xhii, xloi < pxhi)
			if mask.any():
				data_gofp[i] = numpy.nanmin([data_gofp[i], (nstats[mask][:,2]).min() * len(nstats)])
	
	gof_avg = curgof
	gof_total = gof_avg * len(data)
	
	# return data, marked
	marked_binned = []
	# plot data
	ymin = 1e300
	ymax = 0
	for (xloi, xhii, ydatai), gofpi in zip(grouped_data, data_gofp):
		best_gof = -numpy.log10(gofpi + 1e-300)
		# 1e3 and 1e6 correspond roughly to 3 sigma and 5 sigma
		c = 'green' if best_gof < 2 else 'orange' if best_gof < 6. else 'red'
		f = 1. / (xhii - xloi)
		y = ydatai * f
		modelrange_low  = scipy.special.gammaincinv(ydatai + 1, 0.1) * f
		modelrange_high = scipy.special.gammaincinv(ydatai + 1, 0.9) * f
		marked_binned.append(dict(
			x=(xloi + xhii)/2., xerr=(xhii - xloi) / 2.,
			y = y,
			yerr = [[max(0, modelrange_high - y)], [max(0, y - modelrange_low)]],
			color=c)
		)
		ymin = min(ymin, modelrange_low)
		ymax = max(ymax, modelrange_high)
	
	return dict(marked_binned = marked_binned, 
		gof_avg=gof_avg, gof_total=gof_total, stats=stats,
		xlim = (xlo[0], xhi[-1]),
		ylim = (ymin, ymax),
	)



"""
Compute poisson-based GOF
"""
import sys, os
import numpy
import scipy.stats, scipy.special
import matplotlib.pyplot as plt

def build(options, k):
	if len(options) == 1:
		o = options[0]
		if k < len(o):
			#print 'yielding kth option', k, o
			yield [o[k]]
	else:
		for i, o in enumerate(options[0]):
			if i > k:
				break
			for p in build(options[1:], k - i):
				yield [o] + p

def gen_choices(cpart, k):
	nchoices = numpy.prod([len(c) for c in cpart])
	print(' gen?', k, len(cpart), nchoices)
	#print [c for c in build(cpart, k) if c is not None]
	probs = sum((numpy.prod(c) for c in build(cpart, k) if c is not None))
	return probs

	if k == 0:
		probs = numpy.prod([cp[0] for cp in cpart])
	elif k == 1:
		for i in range(len(cpart)):
			probs += numpy.prod([cp[1] if j == i else cp[0] for j, cp in enumerate(cpart)])
	elif k == 2:
		for i in range(len(cpart)):
			probs += numpy.prod([cp[2] if j == i else cp[0] for j, cp in enumerate(cpart)])
			for j in range(i + 1, len(cpart)):
				probs += numpy.prod([cp[1] if k == i or k == j else cp[0] 
					for k, cp in enumerate(cpart)])
	else:
		probs = sum((numpy.prod(c) for c in build(cpart, k)))
		
		for c in range(nchoices):
			prob = 1
			kc = 0
			for cp in cpart:
				ci = c % len(cp)
				kc += ci
				if kc > k:
					break
				c /= len(cp)
				prob *= cp[ci]
			if kc != k:
				continue
			print(' gen!', kc, prob)
			probs += prob
	return probs

def calc_models_range(data):
	stats = []
	lowmodel = numpy.zeros_like(data)*1e-10
	highmodel = numpy.ones_like(data)*1e10
	for i in range(20):
		n = 4**i
		m = len(data) / n
		if n > len(data):
			break
		kplusones = []
		masks = []
		for j in range(int(numpy.ceil(len(data) * 1. / n))):
			mask = slice(j * n, (j + 1) * n)
			dpart = data[mask]
			lowpart  = lowmodel[mask]
			highpart = highmodel[mask]
			
			k = int(dpart.sum()) if len(dpart) > 0 else 0
			kplusones.append(k + 1)
			masks.append(mask)
		
		#print 'GOF MODEL RANGE', k, n, m, 0.1 / m
		lowmodel_parts  = scipy.special.gammaincinv(kplusones, 0.1 / m) / n
		highmodel_parts = scipy.special.gammaincinv(kplusones, 1 - 0.1 / m) / n
		assert not numpy.any(numpy.isnan(lowmodel_parts)), (lowmodel_parts, kplusones, m, n)
		assert not numpy.any(numpy.isnan(highmodel_parts)), (highmodel_parts, kplusones, m, n)
		for mask, lowmodel_part, highmodel_part in zip(masks, lowmodel_parts, highmodel_parts):
			lowpart  = lowmodel[mask]
			highpart = highmodel[mask]
			lowpart [lowpart  < lowmodel_part]  = lowmodel_part
			highpart[highpart > highmodel_part] = highmodel_part
	
	return lowmodel, highmodel

def calc_multigof(data, model):
	#choice_limit = 1e-2 / len(data)
	#choices = numpy.array([[scipy.stats.poisson(m).pmf(i)
			#for i in range(
				#int(scipy.stats.poisson(m).ppf(choice_limit)),
				#int(scipy.stats.poisson(m).ppf(1 - choice_limit) + 1))]
					#for m in model])
	
	stats = []
	for i in range(20):
		n = 4**i
		if n > len(data):
			break
		
		for j in range(int(numpy.ceil(len(data) * 1. / n))):
			dpart = data [j*n:(j + 1)*n]
			mpart = model[j*n:(j + 1)*n]
			#cpart = choices[j*n: (j + 1)*n]
			#print data.shape, dpart.shape, j*n, (j + 1)*n, dpart, data
			k = int(dpart.sum()) if len(dpart) > 0 else 0
			m = mpart.sum()
			
			# compare this probability to all the other k
			#probs = gen_choices(cpart, k)
			stats.append([n, j, 0., m, k]) #  * len(data) / n])
			#print '  multigof', n, j, probs, len(stats)
			# lam = sum(mpart) WRONG!
			# go through all possibilities to get k
	stats = numpy.array(stats)
	m = stats[:,3]
	k = stats[:,4]
	probs = numpy.where(m > 0, scipy.stats.poisson(m).pmf(k), 
		numpy.where(k == 0, 1, 1e-10))
	assert not numpy.isnan(probs).any(), [m, k]
	stats[:,2] = probs
	return stats

def group_adapt(data, nmin = 10):
	i = 0
	while i < len(data):
		#print ' ',i,'--',len(data)
		for j in range(i + 1, len(data) + 1):
			#print ' ',i,j
			if sum(data[i:j + 1]) >= nmin or j + 1 >= len(data):
				yield (i,j + 1)
				#print '  groups', i,j
				break
		i = j + 1

def write_multigof(filename, data, model, skip_plot_ifeq=None):
	segments = list(group_adapt(data))
	stats    = calc_multigof(data, model)
	#gof = -2 * numpy.log10(stats[:,2]).mean()
	#print gof,
	#gof = - numpy.mean([numpy.log10(stats[stats[:,0] == n][:,2]).mean() for n in sorted(set(stats[:,0]))])
	#gof = - 2 * numpy.log(
	#	numpy.mean([exp(numpy.log(stats[stats[:,0] == n][:,2]).sum()) * (stats[:,0] == n).sum() 
	#		for n in sorted(set(stats[:,0]))]))
	gof = -numpy.log10(
		numpy.min([stats[stats[:,0] == n][:,2].min() * (stats[:,0] == n).sum() 
			for n in sorted(set(stats[:,0]))]))
	if skip_plot_ifeq is not None and numpy.abs(gof - skip_plot_ifeq) < 0.01 and os.path.exists(filename + ".pdf"):
		return gof
	
	plt.figure(figsize=(4*1.5,7*1.5))
	plt.subplot(4, 1, 1)
	segdata = [model[i:j].sum() for i, j in segments]
	segerr  = [[s - scipy.stats.poisson(s).ppf(0.1) for s in segdata],
		   [scipy.stats.poisson(s).ppf(0.9) - s for s in segdata]]
	plt.errorbar(x=[(i+j)/2 for i,j in segments],
		xerr=[(j-i)/2 for i,j in segments],
		y=segdata,
		yerr=segerr,
		label='model')
	segmodel = [data[i:j].sum() for i, j in segments]
	segerr  = [[s - scipy.special.gammaincinv(s + 1, 0.1) for s in segmodel],
		   [scipy.special.gammaincinv(s + 1, 0.9) - s for s in segmodel]]
	plt.errorbar(x=[(i+j)/2. for i,j in segments],
		xerr=[(j-i)/2. for i,j in segments],
		y=segdata,
		yerr=segerr,
		label='data')
	plt.legend(loc='best', prop=dict(size=6))
	
	# plot probabilities
	plt.subplot(4, 1, 3)
	colors = dict(list(zip(sorted(set(stats[:,0])), ['b','g','r','c','m','y','k','w'])))
	for n in sorted(set(stats[:,0])):
		j = stats[stats[:,0] == n][:,1] * n
		p = stats[stats[:,0] == n][:,2]
		#print j, p
		plt.plot(j, p, '-x', drawstyle='steps-pre', label='%d' % n,
			color=colors[n])
	plt.gca().set_yscale('log')
	plt.ylim(1e-9, 1)
	plt.legend(loc='best', prop=dict(size=6))
	
	#for l in sorted(stats, key=lambda l: l[2])[:10]:
	#	print l
	
	plt.subplot(4, 1, 2)
	for n in sorted(set(stats[:,0])):
		j = stats[stats[:,0] == n][:,1] * n
		p = stats[stats[:,0] == n][:,3]
		plt.plot(j, p, '--o', drawstyle='steps-pre', label='%d model' % n,
			color=colors[n], markersize=3)
		p = stats[stats[:,0] == n][:,4]
		plt.plot(j, p, '-x', drawstyle='steps-pre', label='%d data' % n,
			color=colors[n])
	plt.gca().set_yscale('log')
	plt.legend(loc='best', prop=dict(size=6))
	
	#print gof
	plt.subplot(4, 1, 4)
	plt.hist(numpy.log10(stats[:,2]), bins=numpy.linspace(-10,0,30),
		label="avg: %.2f" % gof)
	plt.legend(loc='best', prop=dict(size=6))
	plt.savefig(filename + ".pdf")
	plt.close()
	return gof
	
gof_limit = 2

def write_gof(prefix, gof_avg, gof_total):
	verdict = ('good' if gof_avg < gof_limit else 'bad')
	print("%.2f %5s %s"  % (gof_avg, verdict, prefix))
	#file(filename + ".out", 'w').write(str(gof) + "\n")
	numpy.savetxt(prefix + ".out", [gof_total, gof_avg])
	with open(prefix + ".summary", 'w') as f:
		f.write(verdict + "\n")

def load_gof(prefix):
	gof_total, gof_avg = numpy.loadtxt(prefix + ".out")
	return gof_total, gof_avg

	
if __name__ == '__main__':
	gofs = []
	#gof_limit = 1.5

	for filename in sys.argv[1:]:
		table = numpy.loadtxt(filename)
		data, model = table[:len(table)/2,0], table[:len(table)/2,1]
		try:
			prev_gof, prev_avg = load_gof(filename)
		except Exception as e:
			print(e)
			prev_gof, prev_avg = None, None
		gof_avg = write_multigof(filename, data[:650], model[:650], skip_plot_ifeq=prev_avg)
		gof = gof_avg * len(data[:650])
		write_gof(filename, gof_avg, gof)
		gofs.append(gof_avg)

	gofs = numpy.array(gofs)
	plt.figure()
	plt.hist(gofs, bins=numpy.linspace(0,5,21) )
	print('good:', (gofs < gof_limit).sum(), (gofs < gof_limit).sum() * 1. / len(gofs))
	plt.ylabel("number of objects")
	plt.xlabel("gof")
	plt.savefig("gof.pdf", bbox_inches='tight')
	plt.close()





import numpy
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors
import warnings

def grayify(colors):
	RGB_weight = [0.299, 0.587, 0.114]
	luminance = (numpy.dot(colors[:, :3] ** 2, RGB_weight))**0.5
	return luminance.reshape((-1,1)) + numpy.zeros((1,3))

good_cmaps = ['Spectral', 'rainbow', 'YlOrRd', 'spectral', 'RdYlBu', 'GnBu']

def create(n, cmap='YlGnBu', avoid_white=True):
	"""
	cmap
	"""
	cm = matplotlib.cm.get_cmap(cmap)
	colors = cm(numpy.linspace(0, 1, cm.N))
	gray = grayify(colors)
	sign = numpy.sign(gray[1:] - gray[:-1])
	if not numpy.all(sign[sign != 0] == sign[sign!=0][0]):
		warnings.warn('This cmap ("%s") is not monotonic in grayscale!' % cmap)
	colors = cm(numpy.linspace(0, 1, n))
	if avoid_white:
		hsv_colors = matplotlib.colors.rgb_to_hsv(colors[:,:3])
		hsv_colors[:,2] = numpy.where(hsv_colors[:,2] > 0.9, 0.9, hsv_colors[:,2])
		#hsv_colors[:,1] = numpy.where(hsv_colors[:,1] < 0.5, 0.5, hsv_colors[:,1])
		colors = matplotlib.colors.hsv_to_rgb(hsv_colors)
	#colors = grayify(colors)
	return colors.tolist()

"""
Similar to :create:, but meant for many categories so that neighbouring colors are different
"""
def for_categories(n, cmap = 'Spectral', avoid_white=True):
	cm = matplotlib.cm.get_cmap(cmap)
	colors = cm(numpy.linspace(0, 1, n))
	colors_order = []
	for i in range(n/3):
		colors_order += [i, i+n/3, i+2*n/3]
	colors = colors[colors_order]
	
	if avoid_white:
		hsv_colors = matplotlib.colors.rgb_to_hsv(colors[:,:3])
		hsv_colors[:,2] = numpy.where(hsv_colors[:,2] > 0.9, 0.9, hsv_colors[:,2])
		colors = matplotlib.colors.hsv_to_rgb(hsv_colors)
	return colors.tolist()




#!/usr/bin/env python
# -*- coding: utf-8 -*- 

"""
BXA (Bayesian X-ray Analysis) for Sherpa

Copyright: Johannes Buchner (C) 2013-2025

Priors
"""
from math import log10, erf
import numpy
from . import invgauss

def create_uniform_prior_for(parameter):
	"""
	Use for location variables (position)
	The uniform prior gives equal weight in non-logarithmic scale.
	
	:param parameter: Parameter to create a prior for. E.g. xspowerlaw.mypowerlaw.PhoIndex
	"""
	spread = (parameter.max - parameter.min)
	low = parameter.min
	return lambda x: x * spread + low

def create_jeffreys_prior_for(parameter):
	"""deprecated name for create_loguniform_prior_for"""
	return create_loguniform_prior_for(parameter)

def create_loguniform_prior_for(parameter):
	"""
	Use for scale variables (order of magnitude)
	The Jeffreys prior gives equal weight to each order of magnitude between the
	minimum and maximum value. Flat in logarithmic scale.
	
	:param parameter: Parameter to create a prior for. E.g. xspowerlaw.mypowerlaw.norm

	It is usually easier to create an ancillary parameter, and link the 
	actual parameter, like so::

		from sherpa.models.parameter import Parameter
		lognorm = Parameter(modelname='mycomponent', name='lognorm', val=-5, min=-4*2, max=0)
		powerlaw.norm = 10**lognorm

	"""
	assert parameter.min > 0 and parameter.max > 0, ('The limits of "%s" (min and max) are not positive.' % parameter.fullname)
	low = log10(parameter.min)
	spread = log10(parameter.max) - log10(parameter.min)
	return lambda x: 10**(x * spread + low)


def create_gaussian_prior_for(parameter, mean, std):
	"""
	Use for informed variables.
	The Gaussian prior weights by a Gaussian in the parameter. If you 
	would like the logarithm of the parameter to be weighted by
	a Gaussian, create a ancillary parameter (see create_jeffreys_prior_for).
	
	:param parameter: Parameter to create a prior for. E.g. xspowerlaw.mypowerlaw.PhoIndex
	:param mean: Mean of the Gaussian
	:param std: Standard deviation of the Gaussian

	"""
	import scipy.stats
	lo = parameter.min
	hi = parameter.max

	rv = scipy.stats.norm(mean, std)
	xlo = rv.cdf(lo)
	xhi = rv.cdf(hi)
	def gauss_transform(x):
		return rv.ppf(x * (xhi - xlo) + xlo)

	return gauss_transform


def prior_from_file(filename, parameter):
	"""
	Read a custom prior distribution from a file.
	The file should have two columns: cumulative probability
	and value, in ascii format. The cumulative probability
	has to be equally spaced and should exclude 0 and 1.
	
	Returns a sherpa parameter, a list with that parameter inside,
	and the prior function.
	
	If the file only constains a single value, that value is returned
	along with two empty lists.
	"""
	dist = numpy.loadtxt(filename)
	if numpy.shape(dist) == ():
		parameter.val = float(dist)
		return float(dist), [], []
	distz = numpy.array(list(dist[:, 1]) + [dist[-1,1]]*2)
	def custom_priorf(x):
		assert x >= 0
		assert x <= 1
		r = numpy.interp(x, dist[:,0], dist[:,1])
		return r
	return parameter, [parameter], [custom_priorf]

def create_prior_function(priors = [], parameters = None):
	"""
	Combine the prior transformations into a single function.

	This assumes factorized (independent) priors.

	:param priors: individual prior transforms to combine into one function.
		If priors is empty, uniform priors are used on all passed parameters
	:param parameters: If priors is empty, specify the list of parameters.
		Uniform priors will be created for them.
	"""

	if priors == []:
		functions = []
		assert parameters is not None, "you need to pass the parameters if you want automatic uniform priors"
		thawedparmins  = [p.min for p in parameters]
		thawedparmaxes = [p.max for p in parameters]
		for low, high in zip(thawedparmins, thawedparmaxes):
			functions.append(lambda x: x * (high - low) + low)
	else:
		functions = priors
    
	def prior_function(cube, ndim, nparams):
		for i in range(ndim):
			cube[i] = functions[i](cube[i])

	return prior_function



from __future__ import print_function
"""
PCA-based background model.
"""
import numpy
import json
import logging
import warnings
import os

if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat
	from sherpa.models.parameter import Parameter
	from sherpa.models import ArithmeticModel, CompositeModel
	from sherpa.astro.ui import *
	from sherpa.astro.instrument import RSPModelNoPHA, RMFModelNoPHA
else:
	# mock objects when sherpa doc is built
	ArithmeticModel = object

from .pca import get_identity_response

logf = logging.getLogger('bxa.Fitter')
logf.setLevel(logging.INFO)
import scipy.stats

# Model
class KDEModel(ArithmeticModel):
	def __init__(self, channels, model, modelname='KDE'):
		self.channels = channels
		self.model = model

		p0 = Parameter(modelname=modelname, name='lognorm', val=1, min=-5, max=20,
			hard_min=-100, hard_max=100)
		super(ArithmeticModel, self).__init__(modelname, pars=[p0])

	def calc(self, p, left, right, *args, **kwargs):
		lognorm = p[0]
		return 10**lognorm * numpy.interp(left, self.channels, self.model)

	def startup(self, *args):
		pass
	def teardown(self, *args):
		pass
	def guess(self, dep, *args, **kwargs):
		pass

def set_kde_background(id, bw_method=None):
	cts = get_bkg(id).counts.astype(int)
	ndata = len(cts)
	samples = []
	for chan, counts in enumerate(cts):
		samples += [chan] * counts
	kde = scipy.stats.gaussian_kde(samples, bw_method=bw_method)
	channels = numpy.arange(ndata)
	density = kde(channels)
	bkgmodel = KDEModel(channels=channels, model=density / density.sum() * cts.sum())

	response = get_identity_response(id)
	convbkgmodel = response(bkgmodel)
	set_bkg_full_model(id, convbkgmodel)
	return get_bkg_model(id)

__all__ = ['KDEModel', 'set_kde_background']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
BXA (Bayesian X-ray Analysis) for Sherpa

Copyright: Johannes Buchner (C) 2013-2025
"""
import os
import numpy
from math import isnan
from ultranest.integrator import ReactiveNestedSampler, resume_from_similar_file
import ultranest.stepsampler
import warnings
from .priors import create_prior_function

if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat


def default_logging():
	import logging
	logging.basicConfig(filename='bxa.log', level=logging.DEBUG)
	logFormatter = logging.Formatter("[%(name)s %(levelname)s]: %(message)s")
	consoleHandler = logging.StreamHandler()
	consoleHandler.setFormatter(logFormatter)
	consoleHandler.setLevel(logging.INFO)
	logging.getLogger().addHandler(consoleHandler)


def distribution_stats(distribution):
	return distribution.mean(axis=0), distribution.std(axis=0)


def photon_flux_histogram(distribution, nbins=None):
	flux_distribution = distribution[:, -2]
	if nbins is None:
		nbins = len(distribution)**0.5
	x, y = numpy.histogram(flux_distribution, bins=nbins)
	return numpy.vstack([y[:-1], y[1:], x]).transpose()


def energy_flux_histogram(distribution, nbins=None):
	flux_distribution = distribution[:, -1]
	if nbins is None:
		nbins = len(flux_distribution)**0.5
	x, y = numpy.histogram(flux_distribution, bins=nbins)
	return numpy.vstack([y[:-1], y[1:], x]).transpose()


class BXASolver(object):
	def __init__(
		self, id=None, otherids=(), prior=None, parameters=None,
		outputfiles_basename='chains/',
		resume_from=None
	):
		"""
		Set up Bayesian analysis with specified parameters+transformations.

		:param id: See the sherpa documentation of calc_stat.
		:param otherids: See the sherpa documentation of calc_stat.
		:param prior: prior function created with create_prior_function.
		:param parameters: List of parameters to analyse.
		:param outputfiles_basename: prefix for output filenames.
		:param resume_from: prefix for output filenames of a previous run with similar posterior from which to resume

		If prior is None, uniform priors are used on the passed parameters.
		If parameters is also None, all thawed parameters are used.
		"""

		self.id = id
		self.otherids = otherids

		self.fit = ui._session._get_fit(self.id, self.otherids)[1]
		if parameters is None:
			parameters = self.fit.model.thawedpars
		if prior is None:
			prior = create_prior_function(id=id, otherids=otherids, parameters=parameters)

		self.prior = prior
		self.parameters = parameters
		self.outputfiles_basename = outputfiles_basename
		self.set_paramnames()
		self.allowed_stats = (Cash, CStat)
		self.ndims = len(parameters)
		self.vectorized = False

		if resume_from is not None:
			self.paramnames, self.log_likelihood, self.prior_transform, self.vectorized = resume_from_similar_file(
				os.path.join(resume_from, 'chains', 'weighted_post_untransformed.txt'),
				self.paramnames, loglike=self.log_likelihood, transform=self.prior_transform,
				vectorized=False,
			)

	def set_paramnames(self, paramnames=None):
		if paramnames is None:
			self.paramnames = [p.fullname for p in self.parameters]
		else:
			self.paramnames = paramnames

	def get_fit(self):
		return self.fit

	def prior_transform(self, cube):
		"""unit cube transformation.

		see https://johannesbuchner.github.io/UltraNest/priors.html#Dependent-priors
		"""
		params = cube.copy()
		self.prior(params, self.ndims, self.ndims)
		return params

	def log_likelihood(self, cube):
		""" returns -0.5 of the fit statistic."""
		try:
			for i, p in enumerate(self.parameters):
				assert not isnan(cube[i]), 'ERROR: parameter %d (%s) to be set to %f' % (i, p.fullname, cube[i])
				p.val = cube[i]
				# print "%s: %f" % (p.fullname,p.val),
			return -0.5 * self.fit.calc_stat()
		except Exception as e:
			print('Exception in log_likelihood function: ', e)
			for i, p in enumerate(self.parameters):
				print('    Parameter %10s: %f --> %f [%f..%f]' % (p.fullname, p.val, cube[i], p.min, p.max))
			raise e

	def run(
		self, sampler_kwargs={'resume': 'overwrite'}, run_kwargs={'Lepsilon': 0.1},
		speed="safe", resume=None, n_live_points=None,
		frac_remain=None, Lepsilon=0.1, evidence_tolerance=None
	):
		"""Run nested sampling with ultranest.

		:sampler_kwargs: arguments passed to ReactiveNestedSampler (see ultranest documentation)
		:run_kwargs: arguments passed to ReactiveNestedSampler.run() (see ultranest documentation)

		The following arguments are also available directly for backward compatibility:

		:param resume: sets sampler_kwargs['resume']='resume' if True, otherwise 'overwrite'
		:param n_live_points: sets run_kwargs['min_num_live_points']
		:param evidence_tolerance: sets run_kwargs['dlogz']
		:param Lepsilon: sets run_kwargs['Lepsilon']
		:param frac_remain: sets run_kwargs['frac_remain']
		"""

		fit = self.fit
		if False and not isinstance(fit.stat, self.allowed_stats):
			raise RuntimeError("Fit statistic must be cash or cstat, not %s" % fit.stat.name)

		if resume is not None:
			sampler_kwargs['resume'] = 'resume' if resume else 'overwrite'
		run_kwargs['Lepsilon'] = run_kwargs.pop('Lepsilon', Lepsilon)
		del Lepsilon
		if evidence_tolerance is not None:
			run_kwargs['dlogz'] = run_kwargs.pop('dlogz', evidence_tolerance)
		del evidence_tolerance
		if frac_remain is not None:
			run_kwargs['frac_remain'] = run_kwargs.pop('frac_remain', frac_remain)
		del frac_remain
		if n_live_points is not None:
			run_kwargs['min_num_live_points'] = run_kwargs.pop('min_num_live_points', n_live_points)
		del n_live_points

		self.sampler = ReactiveNestedSampler(
			self.paramnames, self.log_likelihood, transform=self.prior_transform,
			log_dir=self.outputfiles_basename,
			vectorized=self.vectorized, **sampler_kwargs)

		if speed == "safe":
			pass
		elif speed == "auto":
			region_filter = run_kwargs.pop('region_filter', True)
			self.sampler.run(max_ncalls=40000, **run_kwargs)

			self.sampler.stepsampler = ultranest.stepsampler.SliceSampler(
				nsteps=1000,
				generate_direction=ultranest.stepsampler.generate_mixture_random_direction,
				adaptive_nsteps='move-distance', region_filter=region_filter
			)
		else:
			self.sampler.stepsampler = ultranest.stepsampler.SliceSampler(
				generate_direction=ultranest.stepsampler.generate_mixture_random_direction,
				nsteps=speed,
				adaptive_nsteps=False,
				region_filter=False)

		self.sampler.run(**run_kwargs)
		self.sampler.print_results()
		self.results = self.sampler.results
		try:
			self.sampler.plot()
		except Exception as e:
			import traceback
			traceback.print_exc()
			warnings.warn("plotting failed.")

		self.set_best_fit()
		return self.results

	def set_best_fit(self):
		"""Sets model to the best fit values."""
		i = numpy.argmax(self.results['weighted_samples']['logl'])
		for p, v in zip(self.parameters, self.results['weighted_samples']['points'][i, :]):
			p.val = v

	def get_distribution_with_fluxes(self, elo=None, ehi=None):
		"""Computes flux posterior samples.

		Returns an array of equally weighted posterior samples (parameter values) with two
		additional columns: the photon fluxes and the energy fluxes.

		The values will be correctly distributed according to the
		analysis run before.
		"""

		r = []
		for row in self.results['samples']:
			for p, v in zip(self.parameters, row):
				p.val = v
			r.append(
				list(row) + [
					ui._session.calc_photon_flux(lo=elo, hi=ehi, id=self.id),
					ui._session.calc_energy_flux(lo=elo, hi=ehi, id=self.id)])
		return numpy.array(r)



"""
This file is not intended to be imported.

Import the sherpa or xspec module instead::

	import bxa.sherpa as bxa

or, for Xspec::

	import bxa.xspec as bxa

"""

__author__ = """Johannes Buchner"""
__email__ = 'johannes.buchner.acad@gmx.com'
__version__ = '5.1.1'



"""
Precomputes a interpolated grid model from a complex model.

That model can then be used for fast access.
"""

import os

if 'MAKESPHINXDOC' not in os.environ:
	from sherpa.astro.ui import *
	from sherpa.models.parameter import Parameter
	from sherpa.models import ArithmeticModel
else:
	ArithmeticModel = object

import itertools
import numpy
from tqdm import tqdm

"""
custom interpolation code, needs
https://github.com/JohannesBuchner/npyinterp
"""
from monointerp import interp


"""
slowmodel:
   a arbitrary model that should be turned into a fast model by gridding
ebins:
   energy bins to compute over, e.g. ebins=numpy.linspace(0.1, 50, 200)
parameters:
   a list of parameters to modify and the number of grid points to use.
   Example:
      parameters = [
        (warmabs.myabs.nH, 21),
        (warmabs.myabs.ionisation, 10),
      ]
      This creates a grid with 21x10 points.
   
   Note that normalisation and redshift should not be part of the gridding,
   they are additionally added to the final model. 
   
   Note that all parameters that are not varied are fixed and should be set to 
   the right (fixed) values.
   
   Note that the interpolation is always linear in the parameters. If you would
   like a interpolation in logarithmic gridding, introduce a helper parameter:
   
     # create helper parameter
     lognH = Parameter(modelname='mymodel', name='nH', val=20, min=20, max=26,
         hard_min=20, hard_max=26)
     # link to the parameter we want to change
     warmabs.myabs.nH = lognH 
     # now pass lognH for logarithmic gridding
     parameters = [
        (lognH, 21),
        (warmabs.myabs.ionisation, 10),
     ]
   
"""
class RebinnedModel(ArithmeticModel):
	def __init__(self, slowmodel, ebins, parameters, filename, modelname='rebinnedmodel'):
		params = [param for param, nbins in parameters]
	
		bins = [numpy.linspace(param.min, param.max, nbins) for param, nbins in parameters]
		left = ebins[:-1]
		right = ebins[1:]
		width = right - left
		ntot = numpy.product([len(bin) for bin in bins])
		try:
			alldata = numpy.load(filename)
			data = alldata['y']
			assert numpy.allclose(alldata['x'], ebins), 'energy binning differs -- plese delete "%s"' % filename
			print('loaded from %s' % filename)
		except IOError:
			print('creating rebinnedmodel, this might take a while')
			print('interpolation setup:')
			print('   energies:', ebins[0], ebins[1], '...', ebins[-2], ebins[-1])
			for (param, nbins), bin in zip(parameters, bins):
				print('   %s: %s - %s with %d points' % (param.fullname, param.min, param.max, nbins))
				print('        ', bin)
		
			data = numpy.zeros((ntot, len(ebins)-1))
			for j, element in enumerate(tqdm(list(itertools.product(*bins)), disable=None)):
				for i, p in enumerate(params):
					if p.val != element[i]:
						p.val = element[i]
				values = [p.val for p in slowmodel.pars]
				model = slowmodel.calc(values, left, right)
				#assert numpy.isfinite(model).all(), ('model:', model)
				#model = modelcum / width
				#assert numpy.isfinite(model).all(), ('model', model)
				data[j] = model
				# (modelcum * width).cumsum()
			print('model created. storing to %s' % filename)
			numpy.savez(filename, x=ebins, y=data)
		self.init(modelname=modelname, x=ebins, data=data, parameters=parameters) 
	
	def init(self, modelname, x, data, parameters):
		#print 'BinReaderModel(%s)' % modelname
		self.data = data
		
		pars = []
		#print '  copying parameters'
		for param, nbins in parameters:
			lo = param.min
			hi = param.max
			newp = Parameter(modelname=modelname, name=param.name, 
				val=param.val, min=lo, max=hi, hard_min=lo, hard_max=hi)
			setattr(self, param.name, newp)
			pars.append(newp)
		#print '  adding norm'
		newp = Parameter(modelname=modelname, name='norm', val=1, min=0, max=1e10, hard_min=-1e300, hard_max=1e300)
		self.norm = newp
		pars.append(newp)
		#print '  adding redshift'
		newp = Parameter(modelname=modelname, name='redshift', val=0, min=0, max=10, hard_min=0, hard_max=100)
		self.redshift = newp
		pars.append(newp)
		
		self.x = x
		self.binnings = [(param.min, param.max, nbins) for param, nbins in parameters]
		super(RebinnedModel, self).__init__(modelname, pars=pars)
	def get(self, coords):
		# compute row to access
		j = 0
		for i, ((lo, hi, n), c) in list(enumerate(zip(self.binnings, coords))):
			k = int((c - lo) * n / (hi - lo))
			if k == n: 
				k = n - 1
			#print '  for param %d: %d/%d -> %d' % (i, k, n, j * n + k)
			j = j * n + k
		#print 'accessing', j
		return self.data[j]
	def calc(self, p, left, right, *args, **kwargs):
		# print('  calc', p, left, right, args, kwargs)
		coords = p[:-2]
		norm = p[-2]
		redshift = p[-1]
		#print '    shifting ...'
		shiftedleft  = left *(1.+redshift)
		shiftedright = right*(1.+redshift)
		x = self.x
		#print '    getting ...'
		y = self.get(coords)
		#print '    y', y
		
		left = x[:-1]
		right = x[1:]
		width = right - left
		assert (y >= 0).all()
		yw = (y).cumsum()
		#for a, b, c, d in zip(x, width, y, yw):
		#	print '%.3f %.3f %.3f %.3f' % (a, b, c, d)
		
		#print '    y*w', y
		#print '    interp ...'
		r = interp(shiftedleft, shiftedright, x, yw)
		#for a, b, c in zip(shiftedleft, shiftedright, r):
		#	print '%.3f %.3f %.3f' % (a, b, c)
		#print '    result', r
		assert numpy.isfinite(r).all(), r
		#print r[len(r)/2], yw[len(yw)/2], norm
		return r * norm



#!/usr/bin/env python
"""
Creates background PCA model from a large sample of background spectra
"""
import astropy.io.fits as pyfits
import numpy
import sys
import h5py
import joblib
import json
import tqdm

from . import compress


def compile(filenames):
	data = []
	instrument = None
	telescope = None

	for filename in filenames:
		#print 'reading', filename
		f = pyfits.open(filename)
		s = f['SPECTRUM']
		y = numpy.copy(s.data['COUNTS'])
		#ncts = y.sum()
		#z = numpy.log10((y + 1.0)/ncts)
		#print ncts, len(y)
		instrument = str(s.header['INSTRUME'])
		telescope = str(s.header['TELESCOP'])
		#if y[:500].mean() > 5:
		if True or y[:200].mean() > 5:
			print('  reading "%s"' % filename)
			data.append(y)
		f.close()

	data = numpy.array(data)

	#with h5py.File(outfile, 'w') as f:
	#	d = f.create_dataset("spectra", data=data, compression="gzip", compression_opts=9, shuffle=True)
	#	d.attrs['INSTRUMENT'] = instrument
	#	d.attrs['TELESCOPE'] = telescope
	return data, instrument, telescope

def repack(filename):
	f = h5py.File(filename, 'r')
	attrs = f['spectra'].attrs
	data = f['spectra'][()]
	mincts = 50 * data.shape[1] # number of counts per bin

	print('original shape: %s' % str(data.shape))
	ncts = data.sum(axis=1)
	selected = ncts>mincts
	print('high-count:    %d have >%d counts' % (selected.sum(), mincts))
	print('count distribution:  %d %d %d %d %d' % tuple(numpy.percentile(ncts, [1, 10, 50, 90, 99])))

	# order by number of counts in the hope we can get spectra in faint and bright background regions together
	indices = numpy.argsort(ncts)
	indices2 = indices[::-1]
	indices3 = numpy.random.randint(0, len(data), 40 * len(data))
	indices = numpy.hstack((indices3, indices2, indices))

	# start with individuals already above threshold
	newdata = []
	initdata = data[selected,:]

	# group together those below threshold, in order of indices, reverse
	pack = 0
	packlen = 0

	for i in tqdm.tqdm(indices[::-1]):
		if ncts[i] > mincts:
			continue
		pack += data[i,:]
		packlen += 1
		if pack.sum() > mincts:
			newdata.append(pack)
			pack = 0
			packlen = 0

	newdata = numpy.vstack(tuple(newdata))
	newdata = numpy.concatenate((initdata, newdata))
	print('repacked shape:', newdata.shape)
	print('dropped:       ', packlen)
	outfile = filename + 'repacked.hdf5'

	with h5py.File(outfile, 'w') as f:
		d = f.create_dataset("spectra", data=newdata, compression="gzip", compression_opts=9, shuffle=True)
		for k, v in attrs.items():
			print('   storing attribute %s = %s' % (k, v))
			d.attrs[k] = v
	return newdata, outfile

def simplify(v):
	try:
		return int(v)
	except:
		return v

def export(outfilename, filename):
	f = h5py.File(filename, 'r')
	#nbins = f['spectra'].shape[1]
	with open(outfilename, 'w') as fout:
		#print("  - getting keys:")
		#for k in list(f.keys()):
		#	print(target, k, f[k].shape)
		data = dict([(k, f[k][()].tolist()) for k in list(f.keys())])
		#print("getting attributes:")
		for k, v in f.attrs.items():
		#	print(target, k, v)
			data[k] = simplify(v)
		json.dump(data, fout, indent=2)


if __name__ == '__main__':
	filenames = sys.argv[1:]
	if len(filenames) == 0:
		sys.stderr.write("""SYNOPSIS: %s <filenames>

Builds a machine-learned background model from a large set of
background spectra (filenames).

Johannes Buchner (C) 2017-2019
""" % sys.argv[0])
		sys.exit(1)

	print("fetching counts from %d files ..." % len(filenames))
	datasets = joblib.Parallel(-1)((joblib.delayed(compile)(filenames[i::10]) for i in range(10)))

	print("combining ...", len(datasets))
	telescopes = set([telescope for _, instrument, telescope in datasets])
	assert len(telescopes) == 1, "trying to combine multiple telescopes!"
	instruments = set([instrument for _, instrument, telescope in datasets])
	instrument = list(instruments)[0]
	telescope = list(telescopes)[0]
	if len(instruments) == 1:
		telescope = telescope + '_' + instrument
	else:
		telescope = telescope

	newdata = numpy.concatenate(tuple([data for data, _, _ in datasets]))
	nbins = newdata.shape[1]
	del datasets
	attrs = {'INSTRUMENT': instrument, 'TELESCOPE': telescope}


	print('combined shape:', newdata.shape)
	outfile = '%s.hdf5' % telescope

	with h5py.File(outfile, 'w') as f:
		d = f.create_dataset("spectra", data=newdata, compression="gzip", compression_opts=9, shuffle=True)
		for k, v in list(attrs.items()):
			print(('   storing attribute %s = %s' % (k, v)))
			d.attrs[k] = v

	newdata, repackedfile = repack(outfile)

	print('applying PCA ...')
	del newdata
	componentfile = compress.run(cmd='create', filename=repackedfile)
	try:
		print('plotting PCA ...')
		compress.run(cmd='components', filename=repackedfile)
	except Exception:
		pass

	print('exporting to "%s.json" ...' % telescope.lower())
	export(telescope.lower() + '_%d.json' % nbins, componentfile)
	print('exporting to "%s.json" ... done' % telescope.lower())



#!/usr/bin/env python
import setuptools
from distutils.core import setup
import warnings

try:
    with open('README.rst') as f:
        long_description = f.read()
except IOError:
    long_description = ''

xspec_available = True
try:
    import xspec
except ImportError:
    xspec_available = False

sherpa_available = True
try:
    import sherpa
except ImportError:
    sherpa_available = False

if not sherpa_available and not xspec_available:
    warnings.warn("BXA is a plugin for xspec/sherpa, but neither xspec nor sherpa are installed in the current environment!")

setup(name='bxa',
    version='5.1.1',
    author='Johannes Buchner',
    url='https://github.com/JohannesBuchner/BXA/',
    author_email='johannes.buchner.acad@gmx.com',
    description='Bayesian X-ray spectral analysis',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    license="GNU General Public License v3",
    long_description=long_description,
    packages=['bxa.xspec', 'bxa.sherpa', 'bxa.sherpa.background', 'bxa', 'autobackgroundmodel', 'bxa.xspec.workflow'],
    install_requires=['ultranest','numpy', 'tqdm', 'corner', 'h5py', 'matplotlib', 'astropy', 'natsort', 'joblib'],
    scripts=['gal.py', 'fixkeywords.py', 'addspec.py', 'bxa_fitbkg.py', 'bxa_fitbkgkde.py'],
    include_package_data=True,
)

if not sherpa_available and not xspec_available:
    warnings.warn("BXA is a plugin for xspec/sherpa, but neither xspec nor sherpa are installed in the current environment!")



# -*- coding: utf-8 -*-
"""

QQ plots and goodness of fit

"""
import os
if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat

import numpy
import json
from math import log10, isnan
from numpy import logical_and

# KS-stat

def KSstat(data, model, staterror=None, syserror=None, weight=None):
	modelc = model.cumsum() / model.sum()
	datac = data.cumsum() / data.sum()
	ks = numpy.abs(modelc - datac).max()
	return ks, 0

def CvMstat(data, model, staterror=None, syserror=None, weight=None):
	modelc = model.cumsum()
	datac = data.cumsum()
	maxmodelc = modelc.max()
	cvm = ((modelc / maxmodelc - datac / datac.max())**2 * model / maxmodelc).sum()
	return cvm, 0

def ADstat(data, model, staterror=None, syserror=None, weight=None):
	modelc = model.cumsum()
	datac = data.cumsum()
	maxmodelc = modelc.max()
	valid = numpy.logical_and(modelc > 0, maxmodelc - modelc > 0)
	modelc = modelc[valid] / maxmodelc
	datac = datac[valid] / datac.max()
	model = model[valid] / maxmodelc
	assert (modelc > 0).all(), ['ADstat has zero cumulative denominator', modelc]
	assert (maxmodelc - modelc > 0).all(), ['ADstat has zero=1-1 cumulative denominator', maxmodelc - modelc]
	ad = ((modelc - datac)**2 / (modelc * (maxmodelc - modelc)) * model).sum()
	return ad, 0

def fake_staterr_func(data):
	return data**0.5

def qq_export(id=None, bkg=False, outfile='qq.txt', elow=None, ehigh=None):
	"""
	Export Q-Q plot into a file for plotting.

	:param id: spectrum id to use (see get_bkg_plot/get_data_plot)
	:param bkg: whether to use get_bkg_plot or get_data_plot
	:param outfile: filename to write results into
	:param elow: low energy limit
	:param ehigh: low energy limit

	Example::

		qq.qq_export('bg', outfile='my_bg_qq', elow=0.2, ehigh=10)

	"""
	# data
	d = ui.get_bkg_plot(id=id) if bkg else ui.get_data_plot(id=id)
	e = d.x
	mask = logical_and(e >= elow, e <= ehigh)
	data = d.y[mask].cumsum()
	d = ui.get_bkg_model_plot(id=id) if bkg else ui.get_model_plot(id=id)
	e = d.xlo
	mask = logical_and(e >= elow, e <= ehigh)
	e = e[mask]
	model = d.y[mask].cumsum()
	last_stat = ui.get_stat()
	ui.set_stat(ksstat)
	ks = ui.calc_stat()
	ui.set_stat(cvmstat)
	cvm = ui.calc_stat()
	ui.set_stat(adstat)
	ad = ui.calc_stat()
	ui.set_stat(last_stat)
	ad = ui.calc_stat()
	
	ui.set_stat('chi2gehrels')
	chi2 = ui.calc_stat()
	ui.set_stat('cstat')
	cstat = ui.calc_stat()
	ui.set_stat(last_stat)
	stats = dict(ks=ks, cvm=cvm, ad=ad, cstat=cstat, chi2=chi2)
	
	numpy.savetxt(outfile, numpy.transpose([e, data, model]))
	json.dump(stats, open(outfile + '.json', 'w'), indent=4)
	
if 'MAKESPHINXDOC' not in os.environ:
	ui.load_user_stat("ksstat", KSstat, fake_staterr_func)
	ui.load_user_stat("cvmstat", CvMstat, fake_staterr_func)
	ui.load_user_stat("adstat", ADstat, fake_staterr_func)
